		/****** Copyright 2014 Thierry Chekouo ****/
		   /*******copying permission************/
/* The program is distributed under the terms of the GNU General Public License*/


import java.util.Arrays;
import java.io.IOException;


public class mainProgram {
// We developed two models and two fitting methods for each model
//1. GPE=the penalized plaid model with the sampling of the penalty parameter, lambda. The model is fitted with a Gibbs sampling.
//2. GPF=the penalized model with a fix value of the penalty parameter lambda anf fitted with  Gibbs sampling procedure
	//2.1 when  lambda=0, we have the plaid model which does not assume any constraint on the amount of overlapping genes and conditions between biclusters
	//2.2 when lambda tends to infinity (lambda>=10^3 is recommended), we assume that biclusters do not overlap as with the Cheng and Church model.
//3. MPE=the penalized model fitted with a Metropolis-Hastings procedure
//4. MPF=the penalized model with a fix values of lambda and fitted a Metropolis-Hastings procedure

//**************** To run the class file with a linux console *************
//java mainProgram "/scratch/biostatistics/tchekouo/Montr/PlaidPenalized/" "data.txt" nbrgenes=400 nbrconditions=50 nbrofbiclusters=2 Gibbs_plaid_lambda sample=1000 burnin=300


///********** Specification of the hyperparameters *******************
static int nu0=1;
static double s20_sigma2=0.05;
static double s20_sigma20=0.05;
static double sigma2alpha=0.5;
static double sigma2beta=0.5;
static double sigma2mu0=0.5;
static double sigma2mu=0.5;

//**** Main method ******
public static void main(String[] args) throws IOException, Exception{
if (args.length<8){
System.out.println("****ENTER AT LEAST 8 INPUTS!!!!!******");
System.exit(1);
}
String represults=args[0];
String datamatrix=args[1];/// data loaded
int n=0;
int p=0;
int K=0;
    try {
n=InitialValuesPenalized.ExtractInt(args[2]); //number of rows of the data matrix
p=InitialValuesPenalized.ExtractInt(args[3]); //number of columns of the data matrix
K=InitialValuesPenalized.ExtractInt(args[4]);//number of bicluster to estimate
    } catch (NumberFormatException e) {
        System.err.println("Arguments:" + args[2]+", "+ args[3] +", "+ args[4]+ " must be an integer.");
        System.exit(1);
    }
if ((K==0)||(n==0)||(p==0)){
System.out.println("You must run with a positive number of biclusters(K), number of genes(p) and conditions(q).");
System.exit(1);
}

double[][] data =new double[n][p];
try {
data=InitialValuesPenalized.ReadArrayText(n,p,datamatrix);
}
catch (IndexOutOfBoundsException e){
System.out.println("The dimensions (p and q) of your data you entered may be invalid.");
}

String model=args[5];
if ((model.equals("MPF"))||(model.equals("GPF"))){
if ((args.length!=9)&&(args.length!=12)){
System.out.println("You have entered "+args.length+" inputs.");
System.out.println("****Enter 9 or 12 inputs for the MPF/GPF Methods!!!!!******");
System.exit(1);}
} else if ((model.equals("MPE"))||(model.equals("GPE"))) {
if ((args.length!=8)&&(args.length!=11)){
System.out.println("You have entered "+args.length+" inputs.");
System.out.println("****Enter 8 or 11 inputs for the MPF/GPF Methods!!!!!******");
System.exit(1);
}
} else {
System.out.println("The model, "+model+", you have entered is invalid. It must be: MPF or GPF or MPE or GPE.");
System.exit(1);
}

int sample=InitialValuesPenalized.ExtractInt(args[6]); //number of samples after the burn in
int burninsample=InitialValuesPenalized.ExtractInt(args[7]); //burn in sample
int K1=K;
double lambda=0;
int[][] rhoSim=new int[n][K1];
int[][] kappaSim=new int[p][K1];
if (args.length==9){
lambda=InitialValuesPenalized.ExtractDouble(args[8]);
} 
else if (args.length==12) {
lambda=InitialValuesPenalized.ExtractDouble(args[8]);
K1=InitialValuesPenalized.ExtractInt(args[9]);//number of known bicluster
String rhoknown=args[10];
String kappaknown=args[11];
try{
 rhoSim=InitialValuesPenalized.RhoInitials(n,K1,rhoknown);//known memberships
 kappaSim=InitialValuesPenalized.RhoInitials(p,K1,kappaknown);//known memberships
} catch (IndexOutOfBoundsException e){
System.out.println("The known membership sizes may be invalid.");
System.exit(1);
}
} else if (args.length==11){ 
K1=InitialValuesPenalized.ExtractInt(args[8]);//number of known bicluster
String rhoknown=args[9];
String kappaknown=args[10];
try{ 
rhoSim=InitialValuesPenalized.RhoInitials(n,K1,rhoknown);//known memberships
 kappaSim=InitialValuesPenalized.RhoInitials(p,K1,kappaknown);//known memberships
} catch (IndexOutOfBoundsException e){
System.out.println("The  known membership sizes may be invalid.");
System.exit(1);
}
}

if (K1==0){
System.out.println("The number of known biclusters is invalid.");
System.exit(1);
}

System.out.println("Running the "+model+" model.");

if ((args.length==9)||(args.length==12)){
System.out.println("Running with lambda = "+lambda);
}

if (args.length>9){
System.out.println("You have entered "+K1+" true biclusters.");
}
System.out.println("Running with "+K+" biclusters.");

double [][] data1=new double[n][p];
double[] lik=new double[sample+burninsample];
double  predict_lik=0;
                int rho_output[][]=new int[n][K];
                int kappa_output[][]=new int[p][K];
                int [][]rho=new int[n][K];
                int [][]kappa=new int[p][K];
                double[]sigma2=new double[K];

                double [][]alpha=new double[n][K];
                double [][]beta=new double[p][K];
 double [][]alpha_max=new double[n][K];
                double [][]beta_max=new double[p][K];                
double [][]alpha1=new double[K][n];
                int [][]rho1=new int[K][n];
                double[] mu=new double[K];
double[][] mu_max=new double[K][1];                
double [][]beta1=new double[K][p];
                int [][]kappa1=new int[K][p];
                double sigma22=0;
                double mu0=0;
                double sigma20=0;
                int NumberSeedInit=1;
                int[][][] RhoKappaInitial=InitialValuesPenalized.RhoKappaInitials(n,p,K, NumberSeedInit,data);
                                for (int i=0;i<n;i++){
                                rho[i]=Arrays.copyOf(RhoKappaInitial[0][i],K);
                                }

                                for (int j=0;j<p;j++){
                                kappa[j]=Arrays.copyOf(RhoKappaInitial[1][j],K);
                                }       
                        for (int k=0;k<K;k++){
                                for (int i=0;i<n;i++){
                                        rho1[k][i]=rho[i][k];
                                        }
                                for(int j=0;j<p;j++){    
                                        kappa1[k][j]=kappa[j][k];
                                }
                                int [] indrho=mcmcMethods.IndicesLabelsNonNuls(rho1[k]);
                                int [] indkappa=mcmcMethods.IndicesLabelsNonNuls(kappa1[k]);
                                if (indrho.length*indkappa.length>0){
                                double[] mualpha =InitialValuesPenalized.muAlphaIni(n,p,data, rho1[k], kappa1[k]);
                                double[] mubeta=InitialValuesPenalized.muBetaIni(n,p,data, rho1[k], kappa1[k]);
                                for (int i=0;i<indrho.length;i++){
                                                alpha1[k][indrho[i]]=mualpha[i];        
                                                alpha[indrho[i]][k]=mualpha[i];
                                        }
                                for (int j=0;j<indkappa.length;j++){
                                        beta1[k][indkappa[j]]=mubeta[j];        
                                        beta[indkappa[j]][k]=mubeta[j];
                                }}
                                
                                 }
                        for (int k=0;k<K;k++){
                mu[k]=InitialValuesPenalized.muInit(n,p,data, rho1[k], kappa1[k]);}
                        mu0=InitialValuesPenalized.mu0Init(n,p,K,data, rho, kappa);
  sigma22=InitialValuesPenalized.sigma2Init(n,p,K,data, alpha, beta, mu, mu0, rho, kappa);   sigma20=InitialValuesPenalized.sigma20InitCC(n,p,K,data, mu0, rho, kappa);
                 for (int k=0;k<K;k++){
  sigma2[k]=InitialValuesPenalized.sigma2InitCC(n,p,data, alpha1[k], beta1[k], mu[k], rho1[k], kappa1[k]);
                                }
                                
                        
                

                double moylambda=0;
                double alpha_lambda=16;
                double beta_lambda=8;
                double[][] moyrho=new double [n][K];
                double[][] moykappa=new double [p][K];
                double[][] moyalpha=new double[n][K];
                double[][] moybeta=new double[p][K];
                double[] moymu=new double[K];
                double moysigma2=0;
                double moysigma20=0;
                double moymu0=0;        
                double[] logPrior=new double[sample+burninsample];  
                double[] LogPosterior=new double[sample+burninsample];
	double maxlik=-Double.MAX_VALUE;
	double maxpost=-Double.MAX_VALUE;
	int indexMaxLik=0;
	int indexMaximumPost=0;
	int kappa_max[][]=new int[p][K];                
	int rho_max[][]=new int[n][K];
	for (int t=0;t<sample+burninsample;t++){  
	    //if ((model.equals("Metropolis_CC"))||(model.equals("Gibbs_CC"))){
if (lambda>=Math.pow(10,3)){
		sigma20=mcmcMethods.sigma20CC(n,p,sigma20,K,data, mu0, rho, kappa,s20_sigma20,nu0);
		mu0=mcmcMethods.mu0(n,p,K,data, sigma20, rho, kappa,sigma2mu0);
		      if (t>=burninsample){moysigma20+=sigma20/sample;}
	     }
	      else{
	       sigma22=mcmcMethods.sigma2Plaid(n,p,K,data, mu,mu0, alpha, beta, rho, kappa,s20_sigma2,nu0);
		    mu0=mcmcMethods.mu0(n,p,K,data, sigma22, rho, kappa,sigma2mu0);
		    if (t>=burninsample){
				moysigma2+=sigma22/sample;
				}}

		    if (t>=burninsample){moymu0+=mu0/sample;}
	      for (int k=0;k<K;k++){
		   data1=mcmcMethods.data1(lambda,n,p,K,k, data, mu, alpha, beta, rho, kappa);
//if ((model.equals("Metropolis_CC"))||(model.equals("Gibbs_CC"))){
if (lambda>=Math.pow(10,3)){
	sigma2[k]=mcmcMethods.sigma2CC(n,p,data1,mu[k], alpha1[k], beta1[k], rho1[k], kappa1[k],s20_sigma2,nu0);
					}
					else {sigma2[k]=sigma22;}
		alpha1[k]=mcmcMethods.alpha(n,p,data1,sigma2[k], rho1[k], kappa1[k],sigma2alpha);
		beta1[k]=mcmcMethods.beta(n,p,data1,sigma2[k], rho1[k], kappa1[k],sigma2beta);
				mu[k]=mcmcMethods.mu(n,p,data1, mu[k],sigma2[k], rho1[k], kappa1[k],sigma2mu);
for (int i=0;i<n;i++){
int rho0=rho1[k][i];
if ((model.equals("MPF"))||(model.equals("MPE"))){
rho1[k][i]=mcmcMethods.proposalRho(model,p,K,i,k,lambda,data1,sigma2[k] ,sigma20,mu0,rho, kappa);
rho1[k][i]=mcmcMethods.AcceptanceRho(p,data1,i, sigma2[k], mu[k], alpha1[k], beta1[k],
							rho1[k][i],rho0, kappa1[k]);}
					else {
			rho1[k][i]=mcmcMethods.GibbsRho(model,n,p,K, i, k, lambda, data1,sigma2[k], sigma20, mu0, mu[k], alpha1[k][i], beta1[k], rho, kappa);
					}
					
					rho[i][k]=rho1[k][i];
					alpha[i][k]=alpha1[k][i];
				
				if (t>=burninsample){
					moyrho[i][k]+=(double)rho[i][k]/sample;
					moyalpha[i][k]+=alpha[i][k]/sample;
					}
					}
	for (int j=0;j<p;j++){
	int kappa0=kappa1[k][j];
	 if ((model.equals("MPF"))||(model.equals("MPE"))){
	kappa1[k][j]=mcmcMethods.proposalKappa(model,n,K,j,k,lambda,data1, sigma2[k],sigma20,mu0, rho,kappa);
	kappa1[k][j]=mcmcMethods.AcceptanceKappa(n,data1,j, sigma2[k], mu[k], alpha1[k], beta1[k],rho1[k],kappa1[k][j], kappa0);}
	else{ kappa1[k][j]=mcmcMethods.GibbsKappa(model,n,p,K, j, k, lambda, data1,sigma2[k], sigma20, mu0, 
								mu[k], alpha1[k], beta1[k][j], rho, kappa);
					}
					
					if (t>=burninsample){moymu[k]+=mu[k]/sample;} 
					kappa[j][k]=kappa1[k][j];
					beta[j][k]=beta1[k][j];
					if (t>=burninsample){
						moykappa[j][k]+=(double)kappa[j][k]/sample;
						moybeta[j][k]+=beta[j][k]/sample;
						}
				}
				}
				if ((model.equals("MPE"))||(model.equals("GPE"))){
				
double lambda_old=lambda;

//if (fixeLambda==0){
if ((model.equals("GPE"))||(model.equals("MPE"))){
double lambda_proposal=mcmcMethods.lambda(n,p,K, alpha_lambda, beta_lambda, rho, kappa);
lambda=mcmcMethods.acceptancelambda(n,p,K,lambda_old,lambda_proposal);
if (t>=burninsample){moylambda+=lambda/sample;}
} 
//else {
//lambda=fixeLambda;}
}   
int divsample=(sample+burninsample)/5;
if (t % divsample==0){System.out.println("Number of MCMC iterations is "+t);}				
//if ((model.equals("Metropolis_CC"))||(model.equals("Gibbs_CC"))){
if (lambda>=Math.pow(10,3)){
lik[t]=mcmcMethods.likelihoodCC(n,p,K,data, rho, kappa, alpha, beta, mu, mu0, sigma2,sigma20);}
	else{lik[t]=mcmcMethods.likelihoodPlaid(n,p,K,data, rho, kappa, alpha, beta, mu, mu0,sigma22); 
	logPrior[t]=mcmcMethods.LogPriorDistribution(n,p,K,data,rho, kappa,alpha,beta, mu, mu0,sigma22,lambda,sigma2alpha,sigma2beta,sigma2mu,sigma2mu0,alpha_lambda,beta_lambda,nu0,s20_sigma2, model);
	}
	LogPosterior[t]=lik[t]+logPrior[t];                    
	 
				if (t>=burninsample){predict_lik+=lik[t]/sample;}
				
			if (t>=burninsample){while(lik[t]>maxlik) { maxlik=lik[t]; indexMaxLik=t;}
			while(LogPosterior[t]>maxpost) { maxpost=LogPosterior[t]; indexMaximumPost=t;}}

			if (t==indexMaximumPost) {
			for (int k=0;k<K;k++){
mu_max[k][0]=mu[k];					
for (int i=0;i<n;i++){
					rho_max[i][k]=rho1[k][i];
					alpha_max[i][k]=alpha[i][k];
					}

				for (int j=0;j<p;j++){
					kappa_max[j][k]=kappa1[k][j];
beta_max[j][k]=beta[j][k];
					}}
			}
	}//end of the t=0 loop

	for (int k=0;k<K;k++){
	for (int i=0;i<n;i++){
	if (moyrho[i][k]>=0.5){rho_output[i][k]=1;
	 }}
	for (int j=0;j<p;j++){
	if (moykappa[j][k]>=0.5){kappa_output[j][k]=1;}}}

			
String 	Namefile="."+model+"."+"K"+String.valueOf(K)+".txt";		
				//double[] MinusLik=new double[sample];
				
				//for (int l=0;l<sample;l++){
				//MinusLik[l]=-lik[l+burninsample];}

double AIC=-2*lik[indexMaxLik]+(n+p-1)*K*2;
double DICc=-4*predict_lik+2*lik[indexMaximumPost];
System.out.println("AIC"+Namefile+"= " +AIC);
System.out.println("DIC Conditional"+Namefile+"= " +DICc);
double meanF1Max=0;         
if (args.length>9){
 meanF1Max=mcmcMethods.meanF1(K,rho_max, kappa_max,K1,rhoSim, kappaSim);
System.out.println("F1"+Namefile+"= " +meanF1Max);
}		
if ((model.equals("MPE"))||(model.equals("GPE"))) {
System.out.println("LambdaEstimate"+Namefile+"= " +moylambda);
}

InitialValuesPenalized.SaveIntArray(represults+"KappaEstimate"+Namefile,kappa_output);
InitialValuesPenalized.SaveIntArray(represults+"RhoEstimate"+Namefile,rho_output);
InitialValuesPenalized.WriteArrayText(represults+"AlphaEstimate"+Namefile,alpha_max);
InitialValuesPenalized.WriteArrayText(represults+"BetaEstimate"+Namefile,beta_max);	

InitialValuesPenalized.WriteArrayText(represults+"MuEstimate"+Namefile,mu_max);
InitialValuesPenalized.saveDouble(represults+"AIC"+Namefile,AIC);
InitialValuesPenalized.saveDouble(represults+"LambdaEstimate"+Namefile,moylambda);
InitialValuesPenalized.saveDouble(represults+"DIC"+Namefile,DICc);
if (args.length>9){
InitialValuesPenalized.saveDouble(represults+"F1"+Namefile,meanF1Max);
}
 System.out.println("End of the MCMC algorithm");
}
}




